import json
from datetime import timedelta
from functools import partial
from unittest import mock
from urllib.parse import parse_qs, urlparse, urlunparse

import pytest
from bs4 import BeautifulSoup
from dateutil.parser import parse as parse_date
from tornado.httputil import url_concat

from jupyterhub import orm, scopes
from jupyterhub.utils import url_path_join, utcnow

from .conftest import new_group_name, new_username
from .utils import add_user, api_request, async_requests, get_page, public_url


@pytest.fixture
def share_user(app):
    """The user to be shared with"""
    yield add_user(app.db, name=new_username("share_with"))


@pytest.fixture
def share(app, user, share_user):
    """Share access to `user`'s default server with `share_user`"""
    db = app.db
    spawner = user.spawner.orm_spawner
    owner = user.orm_user
    filter_ = f"server={owner.name}/{spawner.name}"
    scopes = [f"access:servers!{filter_}"]
    yield orm.Share.grant(db, spawner, share_user, scopes=scopes)


@pytest.fixture
def group_share(app, user, group, share_user):
    """Share with share_user via group membership"""
    db = app.db
    app.db.commit()
    spawner = user.spawner.orm_spawner
    owner = user.orm_user
    filter_ = f"server={owner.name}/{spawner.name}"
    scopes = [f"read:servers!{filter_}"]
    group.users.append(share_user)
    yield orm.Share.grant(db, spawner, group, scopes=scopes)


@pytest.fixture
def populate_shares(app, user, group, share_user):
    group_a = orm.Group(name=new_group_name("a"))
    group_b = orm.Group(name=new_group_name("b"))
    group_c = orm.Group(name=new_group_name("c"))
    app.db.add(group_a)
    app.db.add(group_b)
    app.db.add(group_c)
    app.db.commit()
    in_a = add_user(app.db, name=new_username("in-a"))
    in_a_b = add_user(app.db, name=new_username("in-a-b"))
    in_b = add_user(app.db, name=new_username("in-b"))
    not_in = add_user(app.db, name=new_username("not-in"))

    group_a.users = [in_a, in_a_b]
    group_b.users = [in_b, in_a_b]
    app.db.commit()

    user_1 = add_user(app.db, name=new_username("server"))
    user_2 = add_user(app.db, name=new_username("server"))
    user_3 = add_user(app.db, name=new_username("server"))
    user_4 = add_user(app.db, name=new_username("server"))

    # group a has access to user_1
    # group b has access to user_2
    # both groups have access to user_3
    # user in-a also has access to user_4
    orm.Share.grant(
        app.db,
        app.users[user_1].spawner.orm_spawner,
        group_a,
    )
    orm.Share.grant(
        app.db,
        app.users[user_2].spawner.orm_spawner,
        group_b,
    )
    orm.Share.grant(
        app.db,
        app.users[user_3].spawner.orm_spawner,
        group_a,
    )
    orm.Share.grant(
        app.db,
        app.users[user_3].spawner.orm_spawner,
        group_b,
    )
    orm.Share.grant(
        app.db,
        app.users[user_4].spawner.orm_spawner,
        in_a,
    )
    orm.Share.grant(
        app.db,
        app.users[user_4].spawner.orm_spawner,
        not_in,
    )

    # return a dict showing who should have access to what
    return {
        "users": {
            in_a.name: [user_1.name, user_3.name, user_4.name],
            in_b.name: [user_2.name, user_3.name],
            # shares are _not_ deduplicated if granted
            # both via user and group
            in_a_b.name: [user_1.name, user_2.name, user_3.name, user_3.name],
            not_in.name: [user_4.name],
        },
        "groups": {
            group_a.name: [user_1.name, user_3.name],
            group_b.name: [user_2.name, user_3.name],
            group_c.name: [],
        },
    }


def expand_scopes(scope_str, user, group=None, share_with=None):
    """utility to expand scopes used in parametrized tests

    Turns "read:users!user=USER,shares!group=SHARE_WITH"
    into
        [
            "read:users!user=username",
            "shares!group=groupname",
        ]
    """
    scopes = []
    replacements = {}

    def _get_name(str_or_obj):
        """Allow a string name or an object with a name attribute

        string names are used for tests where something doesn't exist
        """
        if isinstance(str_or_obj, str):
            return str_or_obj
        else:
            return str_or_obj.name

    username = _get_name(user)
    replacements["USER"] = username
    replacements["SERVER"] = username + "/"
    if group:
        replacements["GROUP"] = _get_name(group)
    if share_with:
        replacements["SHARE_WITH"] = _get_name(share_with)
    for scope in scope_str.split(","):
        for a, b in replacements.items():
            scope = scope.replace(a, b)
        scopes.append(scope)
    return scopes


@pytest.mark.parametrize(
    "share_with", ["user", pytest.param("group", id="share_with=group")]
)
def test_create_share(app, user, share_user, group, share_with):
    db = app.db
    spawner = user.spawner.orm_spawner
    owner = user.orm_user
    share_attr = share_with
    if share_with == "group":
        share_with = group
    elif share_with == "user":
        share_with = share_user
    scopes = [f"access:servers!server={owner.name}/{spawner.name}"]
    before = orm.Share.now()
    share = orm.Share.grant(db, spawner, share_with, scopes=scopes)
    after = orm.Share.now()
    assert share.scopes == scopes
    assert share.owner is owner
    assert share.spawner is spawner
    assert getattr(share, share_attr) is share_with
    assert share.created_at
    assert before <= share.created_at <= after
    assert share in share_with.shared_with_me
    assert share in spawner.shares
    assert share in owner.shares
    if share_attr == 'user':
        assert share not in share_with.shares
    assert share not in owner.shared_with_me
    # compute repr for coverage
    repr(share)
    db.delete(share_with)
    db.commit()


def test_create_share_bad(app, user, share_user, mockservice):
    db = app.db
    service = mockservice
    spawner = user.spawner.orm_spawner
    owner = user.orm_user
    scopes = [f"access:servers!server={owner.name}/{spawner.name}"]
    with pytest.raises(ValueError):
        orm.Share.grant(db, spawner, share_user, scopes=[])
    with pytest.raises(TypeError):
        orm.Share.grant(db, spawner, service, scopes=scopes)


def test_update_share(app, share):
    db = app.db
    # shift into past
    share.created_at -= timedelta(hours=1)
    created_at = share.created_at
    db.commit()

    # grant additional scopes
    filter = f"server={share.owner.name}/{share.spawner.name}"
    more_scopes = [f"read:servers!{filter}"]
    all_scopes = sorted(share.scopes + more_scopes)

    share2 = orm.Share.grant(db, share.spawner, share.user, scopes=more_scopes)
    assert share2 is share
    assert share.created_at == created_at
    assert share.scopes == all_scopes

    # fully overlapping scopes
    share3 = orm.Share.grant(db, share.spawner, share.user, scopes=share.scopes[:1])
    assert share3 is share
    assert share.created_at == created_at
    assert share.scopes == all_scopes

    # revoke scopes not held
    share4 = orm.Share.revoke(
        db,
        share.spawner,
        share.user,
        scopes=[f"admin:servers!{filter}"],
    )
    assert share4 is share
    assert share.created_at == created_at
    assert share.scopes == all_scopes

    # revoke some scopes
    share5 = orm.Share.revoke(
        db,
        share.spawner,
        share.user,
        scopes=all_scopes[:1],
    )
    remaining_scopes = all_scopes[1:]
    assert share5 is share
    assert share.created_at == created_at
    assert share.scopes == remaining_scopes

    # revoke remaining scopes
    share5 = orm.Share.revoke(
        db,
        share.spawner,
        share.user,
        scopes=remaining_scopes,
    )
    assert share5 is None
    found_share = orm.Share.find(db, spawner=share.spawner, share_with=share.user)
    assert found_share is None


@pytest.mark.parametrize(
    "to_delete",
    [
        "owner",
        "spawner",
        "share_with_user",
        "share_with_group",
    ],
)
def test_share_delete_cascade(to_delete, app, user, group):
    db = app.db
    if "group" in to_delete:
        share_with = group
    else:
        share_with = add_user(db, app, name=new_username("share_with")).orm_user
    spawner = user.spawner.orm_spawner
    owner = user.orm_user
    scopes = [f"access:servers!server={owner.name}/{spawner.name}"]
    assert spawner.name is not None
    assert spawner.user.name
    assert share_with.name
    share = orm.Share.grant(db, spawner, share_with, scopes=scopes)
    assert share in share_with.shared_with_me
    share_id = share.id
    if to_delete == "owner":
        app.users.delete(owner)
        assert share_with.shared_with_me == []
    elif to_delete == "spawner":
        # pass
        db.delete(spawner)
        user.spawners.pop(spawner.name)
        db.commit()
        assert owner.shares == []
        assert share_with.shared_with_me == []
    elif to_delete == "share_with_user":
        app.users.delete(share_with)
        assert owner.shares == []
        assert spawner.shares == []
    elif to_delete == "share_with_group":
        db.delete(share_with)
        db.commit()
        assert owner.shares == []
        assert spawner.shares == []
    else:
        raise ValueError(f"unexpected {to_delete=}")
    # make sure it's gone
    assert db.query(orm.Share).filter_by(id=share_id).one_or_none() is None


def test_share_scopes(app, share_user, share):
    db = app.db
    user_scopes = scopes.get_scopes_for(share_user)
    assert set(share.scopes).issubset(user_scopes)
    # delete share, no longer have scopes
    db.delete(share)
    db.commit()
    user_scopes = scopes.get_scopes_for(share_user)
    assert not set(share.scopes).intersection(user_scopes)


def test_share_group_scopes(app, share_user, group_share):
    # make sure share is actually in the group (make sure group_share fixture worked)
    db = app.db
    share = group_share
    assert group_share.group in share_user.groups
    user_scopes = scopes.get_scopes_for(share_user)
    assert set(share.scopes).issubset(user_scopes)
    # delete share, no longer have scopes
    db.delete(share)
    db.commit()
    user_scopes = scopes.get_scopes_for(share_user)
    assert not set(share.scopes).intersection(user_scopes)


def test_share_code(app, user, share_user):
    spawner = user.spawner.orm_spawner
    user = spawner.user
    code_scopes = sorted(
        [
            f"read:servers!server={user.name}/",
            f"access:servers!server={user.name}/",
        ]
    )
    orm_code, code = orm.ShareCode.new(
        app.db,
        spawner,
        scopes=code_scopes,
    )
    assert sorted(orm_code.scopes) == code_scopes
    assert orm_code.owner is user
    assert orm_code.spawner is spawner
    assert orm_code in spawner.share_codes
    assert orm_code in user.share_codes

    share_with_scopes = scopes.get_scopes_for(share_user)
    for scope in code_scopes:
        assert scope not in share_with_scopes

    assert orm_code.exchange_count == 0
    assert orm_code.last_exchanged_at is None
    # do it twice, shouldn't change anything
    orm_code.exchange(share_user)
    assert orm_code.exchange_count == 1
    assert orm_code.last_exchanged_at is not None
    now = orm_code.now()
    assert now - timedelta(10) <= orm_code.last_exchanged_at <= now + timedelta(10)

    share_with_scopes = scopes.get_scopes_for(share_user)
    for scope in code_scopes:
        assert scope in share_with_scopes


def test_share_code_expires(app, user, share_user):
    db = app.db
    spawner = user.spawner.orm_spawner
    user = spawner.user
    orm_code, code = orm.ShareCode.new(
        db,
        spawner,
        scopes=[
            f"access:servers!server={user.name}/",
        ],
    )
    # check expiration
    assert orm_code.expires_at
    now = orm_code.now()
    assert (
        now - timedelta(10)
        <= orm_code.expires_at
        <= now + timedelta(seconds=orm.ShareCode.default_expires_in + 10)
    )
    orm.ShareCode.purge_expired(db)
    found = orm.ShareCode.find(db, code=code)
    assert found
    assert found.id == orm_code.id

    with mock.patch(
        'jupyterhub.orm.ShareCode.now', staticmethod(lambda: now + timedelta(hours=1))
    ):
        orm.ShareCode.purge_expired(db)
        found = orm.ShareCode.find(db, code=code)
        assert found
        assert found.id == orm_code.id

    with mock.patch(
        'jupyterhub.orm.ShareCode.now', staticmethod(lambda: now + timedelta(hours=25))
    ):
        found = orm.ShareCode.find(db, code=code)
        assert found is None
        # try exchanging expired code
        with pytest.raises(ValueError):
            orm_code.exchange(share_user)

    # expired code, should have been deleted
    found = orm.ShareCode.find(db, code=code)
    assert found is None
    assert db.query(orm.ShareCode).filter_by(id=orm_code.id).one_or_none() is None


# API tests


@pytest.mark.parametrize(
    "kind",
    [
        ("user"),
        (pytest.param("group", id="kind=group")),
    ],
)
async def test_shares_api_user_group_doesnt_exist(
    app,
    user,
    group,
    share_user,
    kind,
):
    # make sure default spawner exists
    spawner = user.spawner  # noqa
    body = {}
    if kind == "user":
        body["user"] = "nosuchuser"
    elif kind == "group":
        body["group"] = "nosuchgroup"

    r = await api_request(
        app, f"/shares/{user.name}/", method="post", data=json.dumps(body)
    )
    assert r.status_code == 400


@pytest.mark.parametrize(
    "which",
    [
        ("user"),
        ("server"),
    ],
)
async def test_shares_api_target_doesnt_exist(
    app,
    user,
    group,
    share_user,
    which,
):
    # make sure default spawner exists
    if which == "server":
        share_path = f"/shares/{user.name}/nosuchserver"
    elif which == "user":
        share_path = "/shares/nosuchuser/"
    body = {"user": share_user.name}

    r = await api_request(app, share_path, method="post", data=json.dumps(body))
    assert r.status_code == 404


@pytest.mark.parametrize(
    "have_scopes, share_scopes, with_user, with_group, status",
    [
        (None, None, True, False, 200),
        ("shares", None, True, False, 403),
        (
            "shares!server=SERVER,servers!server=SERVER,read:users:name!user=SHARE_WITH",
            "read:servers!server=SERVER,access:servers!server=SERVER",
            True,
            False,
            200,
        ),
        (None, "read:servers!server=other/", False, True, 400),
        (
            "shares,access:servers,read:users:name",
            "admin:servers!server=SERVER",
            False,
            True,
            403,
        ),
        (None, None, False, False, 400),
        (None, None, "nosuchuser", False, 400),
        (None, None, False, "nosuchgroup", 400),
        (None, None, True, True, 400),
        (None, None, True, False, 200),
    ],
)
async def test_shares_api_create(
    app,
    user,
    group,
    share_user,
    create_user_with_scopes,
    have_scopes,
    share_scopes,
    with_user,
    with_group,
    status,
):
    # make sure default spawner exists
    spawner = user.spawner  # noqa
    body = {}
    share_with = share_user
    if with_user:
        body["user"] = share_user.name if with_user == True else with_user
    if with_group:
        body["group"] = group.name if with_group == True else with_group
        share_with = group
    _expand_scopes = partial(expand_scopes, user=user, share_with=share_with)

    expected_scopes = _expand_scopes("access:servers!server=SERVER")
    if share_scopes:
        share_scopes = _expand_scopes(share_scopes)
        expected_scopes.extend(share_scopes)
        body["scopes"] = share_scopes

    expected_scopes = sorted(set(expected_scopes))

    if have_scopes is None:
        # default: needed permissions
        have_scopes = "shares,read:users:name,read:groups:name"

    requester = create_user_with_scopes(*_expand_scopes(have_scopes))

    r = await api_request(
        app,
        f"/shares/{user.name}/",
        method="post",
        data=json.dumps(body),
        name=requester.name,
    )
    assert r.status_code == status
    if r.status_code < 300:
        share_model = r.json()
        assert "scopes" in share_model
        assert sorted(share_model["scopes"]) == expected_scopes


@pytest.mark.parametrize(
    "have_scopes, before_scopes, revoke_scopes, after_scopes, with_user, with_group, status",
    [
        ("read:shares", None, None, None, True, False, 403),
        ("shares", None, None, None, True, False, 200),
        ("shares!user=USER", None, None, None, False, True, 200),
        (None, "read:servers!server=SERVER", None, None, True, False, 200),
        (
            None,
            "access:servers!server=SERVER",
            "read:servers!server=SERVER",
            "access:servers!server=SERVER",
            True,
            False,
            200,
        ),
        (None, None, None, None, "nosuchuser", False, 200),
        (None, None, None, None, False, "nosuchgroup", 200),
        (None, None, None, None, False, False, 400),
        (None, None, None, None, True, True, 400),
    ],
)
async def test_shares_api_revoke(
    app,
    user,
    group,
    share_user,
    create_user_with_scopes,
    have_scopes,
    before_scopes,
    revoke_scopes,
    after_scopes,
    with_user,
    with_group,
    status,
):
    db = app.db
    # make sure default spawner exists
    spawner = user.spawner.orm_spawner  # noqa
    body = {}
    share_with = share_user
    if with_user:
        body["user"] = share_user.name if with_user == True else with_user
    if with_group:
        body["group"] = group.name if with_group == True else with_group
        share_with = group
    _expand_scopes = partial(expand_scopes, user=user, share_with=share_with)

    if revoke_scopes:
        revoke_scopes = _expand_scopes(revoke_scopes)
        body["scopes"] = revoke_scopes

    if after_scopes:
        after_scopes = _expand_scopes(after_scopes)

    if before_scopes:
        orm.Share.grant(db, spawner, share_with, scopes=_expand_scopes(before_scopes))

    if have_scopes is None:
        # default: needed permissions
        have_scopes = "shares,read:users:name,read:groups:name"

    requester = create_user_with_scopes(*_expand_scopes(have_scopes))

    r = await api_request(
        app,
        f"/shares/{user.name}/",
        method="patch",
        data=json.dumps(body),
        name=requester.name,
    )
    assert r.status_code == status
    if r.status_code < 300:
        share_model = r.json()
        if not after_scopes:
            # no scopes specified, full revocation
            assert share_model == {}
            return
        assert share_model["scopes"] == after_scopes


@pytest.mark.parametrize(
    "have_scopes, status",
    [
        ("shares", 204),
        ("shares!user=USER", 204),
        ("shares!server=SERVER", 204),
        ("read:shares", 403),
        ("shares!server=USER/other", 404),
        ("shares!user=other", 404),
    ],
)
async def test_shares_api_revoke_all(
    app,
    user,
    group,
    share_user,
    create_user_with_scopes,
    have_scopes,
    status,
):
    db = app.db
    # make sure default spawner exists
    spawner = user.spawner.orm_spawner  # noqa
    orm.Share.grant(db, spawner, share_user)
    orm.Share.grant(db, spawner, group)
    _expand_scopes = partial(expand_scopes, user=user)

    if have_scopes is None:
        # default: needed permissions
        have_scopes = "shares"

    requester = create_user_with_scopes(*_expand_scopes(have_scopes))

    r = await api_request(
        app,
        f"/shares/{user.name}/",
        method="delete",
        name=requester.name,
    )
    assert r.status_code == status

    # get updated share list
    r = await api_request(
        app,
        f"/shares/{user.name}/",
    )
    share_list = r.json()

    if status >= 400:
        assert len(share_list["items"]) == 2
    else:
        assert len(share_list["items"]) == 0


@pytest.mark.parametrize(
    "kind, case",
    [
        ("users", "in-a"),
        ("users", "in-b"),
        ("users", "in-a-b"),
        ("users", "not-in"),
        ("users", "notfound"),
        ("groups", "a"),
        ("groups", "b"),
        ("groups", "c"),
        ("groups", "notfound"),
    ],
)
async def test_shared_api_list_user_group(
    app, populate_shares, create_user_with_scopes, kind, case
):
    if case == "notfound":
        name = "notfound"
    else:
        # find exact name, which will look like `{kind}-123`
        for name, server_names in populate_shares[kind].items():
            if name.rpartition("-")[0] == case:
                break
        else:
            raise ValueError(f"Did not find {case} in {populate_shares[kind].keys()}")

    r = await api_request(app, f"/{kind}/{name}/shared")
    if name == "notfound":
        assert r.status_code == 404
        return
    else:
        assert r.status_code == 200
    shares = r.json()
    found_shares = sorted(
        [share["server"]["user"]["name"] for share in shares["items"]]
    )
    expected_shares = sorted(server_names)
    assert found_shares == expected_shares


@pytest.mark.parametrize(
    "kind, have_scopes, get_status, delete_status",
    [
        ("users", "users", 403, 403),
        ("users", "read:users", 403, 403),
        ("users", "read:users!user=USER", 403, 403),
        ("users", "read:users:shares!user=SHARE_WITH", 200, 403),
        ("users", "users:shares!user=SHARE_WITH", 200, 204),
        ("users", "users:shares!user=other", 404, 404),
        ("groups", "groups", 403, 403),
        ("groups", "read:groups", 403, 403),
        ("groups", "read:groups!group=group", 403, 403),
        ("groups", "read:groups:shares!group=SHARE_WITH", 200, 403),
        ("groups", "groups:shares!group=SHARE_WITH", 200, 204),
        ("groups", "groups:shares!group=other", 404, 404),
    ],
)
async def test_single_shared_api(
    app,
    user,
    share_user,
    group,
    create_user_with_scopes,
    kind,
    have_scopes,
    get_status,
    delete_status,
):
    db = app.db
    share_user.groups.append(group)
    db.commit()
    spawner = user.spawner.orm_spawner

    if kind == "users":
        share_with = share_user
    else:
        share_with = group

    _expand_scopes = partial(expand_scopes, user=user, share_with=share_with)

    requester = create_user_with_scopes(*_expand_scopes(have_scopes))

    share_scopes = [f"access:servers!server={user.name}/"]
    share = orm.Share.grant(
        db, spawner=spawner, share_with=share_with, scopes=share_scopes
    )
    api_url = f"/{kind}/{share_with.name}/shared/{user.name}/"

    fetch_share = partial(api_request, app, api_url, name=requester.name)
    r = await fetch_share()
    assert r.status_code == get_status
    if get_status < 300:
        # check content
        pass

    r = await fetch_share(method="delete")
    assert r.status_code == delete_status
    if delete_status < 300:
        assert r.text == ""
    else:
        # manual delete
        db.delete(share)
        db.commit()

    # now share doesn't exist, should 404
    assert orm.Share.find(db, spawner=spawner, share_with=share_with) is None

    r = await fetch_share()
    assert r.status_code == 404 if get_status < 300 else get_status
    r = await fetch_share(method="delete")
    assert r.status_code == 404 if delete_status < 300 else delete_status


@pytest.mark.parametrize(
    "kind, have_scopes, get_status, delete_status",
    [
        ("users", "read:users:shares!user=SHARE_WITH", 404, 403),
        ("users", "users:shares!user=SHARE_WITH", 404, 404),
        ("users", "users:shares!user=other", 404, 404),
        ("groups", "read:groups:shares!group=SHARE_WITH", 404, 403),
        ("groups", "groups:shares!group=SHARE_WITH", 404, 404),
        ("groups", "groups:shares!group=other", 404, 404),
    ],
)
async def test_single_shared_api_no_such_owner(
    app,
    user,
    share_user,
    group,
    create_user_with_scopes,
    kind,
    have_scopes,
    get_status,
    delete_status,
):
    db = app.db
    share_user.groups.append(group)
    db.commit()
    spawner = user.spawner.orm_spawner  # noqa

    if kind == "users":
        share_with = share_user
    else:
        share_with = group

    owner_name = "nosuchname"

    _expand_scopes = partial(expand_scopes, user=owner_name, share_with=share_with)

    requester = create_user_with_scopes(*_expand_scopes(have_scopes))

    api_url = f"/{kind}/{share_with.name}/shared/{owner_name}/"

    fetch_share = partial(api_request, app, api_url, name=requester.name)
    r = await fetch_share()
    assert r.status_code == get_status

    r = await fetch_share(method="delete")
    assert r.status_code == delete_status


@pytest.mark.parametrize(
    "kind",
    [
        ("users"),
        ("groups"),
    ],
)
async def test_single_shared_api_no_such_target(
    app, user, share_user, group, create_user_with_scopes, kind
):
    db = app.db
    share_user.groups.append(group)
    db.commit()
    spawner = user.spawner.orm_spawner  # noqa
    share_with = "nosuch" + kind

    requester = create_user_with_scopes(f"{kind}:shares")

    api_url = f"/{kind}/{share_with}/shared/{user.name}/"

    fetch_share = partial(api_request, app, api_url, name=requester.name)
    r = await fetch_share()
    assert r.status_code == 404

    r = await fetch_share(method="delete")
    assert r.status_code == 404


@pytest.mark.parametrize(
    "have_scopes, n_groups, n_users, ok",
    [
        ("shares", 0, 0, True),
        ("read:shares", 0, 2, True),
        ("read:shares!user=USER", 3, 0, True),
        ("read:shares!server=SERVER", 2, 1, True),
        ("read:users:shares", 0, 0, False),
    ],
)
async def test_shares_api_list_server(
    app, user, share_user, create_user_with_scopes, have_scopes, n_groups, n_users, ok
):
    db = app.db
    spawner = user.spawner.orm_spawner

    _expand_scopes = partial(expand_scopes, user=user, share_with=share_user)

    requester = create_user_with_scopes(*_expand_scopes(have_scopes))

    expected_shares = []
    for i in range(n_users):
        u = create_user_with_scopes().orm_user
        orm.Share.grant(db, spawner, u)
        expected_shares.append(f"user:{u.name}")

    for i in range(n_groups):
        group = orm.Group(name=new_group_name())
        db.add(group)
        db.commit()
        orm.Share.grant(db, spawner, group)
        expected_shares.append(f"group:{group.name}")
    expected_shares = sorted(expected_shares)
    r = await api_request(app, f"/shares/{user.name}/", name=requester.name)
    if ok:
        assert r.status_code == 200
    else:
        assert r.status_code == 403
        return
    shares = r.json()
    found_shares = []
    for share in shares["items"]:
        assert share["user"] or share["group"]
        if share["user"]:
            found_shares.append(f"user:{share['user']['name']}")
        elif share["group"]:
            found_shares.append(f"group:{share['group']['name']}")
    found_shares = sorted(found_shares)
    assert found_shares == expected_shares


@pytest.mark.parametrize(
    "have_scopes, n_groups, n_users, status",
    [
        ("shares", 0, 0, 200),
        ("read:shares", 0, 2, 200),
        ("read:shares!user=USER", 3, 0, 200),
        ("read:shares!user=other", 3, 0, 404),
        ("read:shares!server=SERVER", 2, 1, 404),
        ("read:users:shares", 0, 0, 403),
    ],
)
async def test_shares_api_list_user(
    app,
    user,
    share_user,
    create_user_with_scopes,
    have_scopes,
    n_groups,
    n_users,
    status,
):
    db = app.db
    spawner = user.spawner.orm_spawner

    _expand_scopes = partial(expand_scopes, user=user, share_with=share_user)

    requester = create_user_with_scopes(*_expand_scopes(have_scopes))

    expected_shares = []
    for i in range(n_users):
        u = create_user_with_scopes().orm_user
        orm.Share.grant(db, spawner, u)
        expected_shares.append(f"user:{u.name}")

    for i in range(n_groups):
        group = orm.Group(name=new_group_name())
        db.add(group)
        db.commit()
        orm.Share.grant(db, spawner, group)
        expected_shares.append(f"group:{group.name}")
    expected_shares = sorted(expected_shares)
    r = await api_request(app, f"/shares/{user.name}", name=requester.name)
    assert r.status_code == status
    if status >= 400:
        return
    shares = r.json()
    found_shares = []
    for share in shares["items"]:
        assert share["user"] or share["group"]
        if share["user"]:
            found_shares.append(f"user:{share['user']['name']}")
        elif share["group"]:
            found_shares.append(f"group:{share['group']['name']}")
    found_shares = sorted(found_shares)
    assert found_shares == expected_shares


async def test_shares_api_list_no_such_owner(app):
    r = await api_request(app, "/shares/nosuchuser")
    assert r.status_code == 404
    r = await api_request(app, "/shares/nosuchuser/")
    assert r.status_code == 404
    r = await api_request(app, "/shares/nosuchuser/namedserver")
    assert r.status_code == 404


@pytest.mark.parametrize(
    "method",
    [
        "post",
        "patch",
        "delete",
    ],
)
async def test_share_api_server_required(app, user, method):
    """test methods defined on /shares/:user/:server not defined on /shares/:user"""
    r = await api_request(app, f"/shares/{user.name}", method=method)
    assert r.status_code == 405


async def test_share_flow_full(
    app, full_spawn, user, share_user, create_user_with_scopes
):
    """Exercise the full process of sharing and then accessing a shared server"""
    user = create_user_with_scopes(
        "shares!user", "self", f"read:users:name!user={share_user.name}"
    )
    # start server
    await user.spawn("")
    await app.proxy.add_user(user)
    spawner = user.spawner

    # grant access
    share_url = f"shares/{user.name}/{spawner.name}"
    r = await api_request(
        app,
        share_url,
        method="post",
        name=user.name,
        data=json.dumps({"user": share_user.name}),
    )
    r.raise_for_status()
    share_model = r.json()

    # attempt to _use_ access
    user_url = public_url(app, user) + "api/contents/"
    token = share_user.new_api_token()
    r = await async_requests.get(user_url, headers={"Authorization": f"Bearer {token}"})
    r.raise_for_status()

    # revoke access
    r = await api_request(
        app,
        share_url,
        method="patch",
        name=user.name,
        data=json.dumps(
            {
                "scopes": share_model["scopes"],
                "user": share_user.name,
            }
        ),
    )
    r.raise_for_status()
    # new request with new token to avoid cache
    token = share_user.new_api_token()
    r = await async_requests.get(user_url, headers={"Authorization": f"Bearer {token}"})
    assert r.status_code == 403


# share codes


@pytest.mark.parametrize(
    "method",
    [
        "post",
        "delete",
    ],
)
async def test_share_codes_api_server_required(app, user, method):
    """test methods defined on /share-codes/:user/:server not defined on /share-codes/:user"""
    r = await api_request(app, f"/share-codes/{user.name}", method=method)
    assert r.status_code == 405


@pytest.mark.parametrize(
    "have_scopes, n_codes, level, status",
    [
        ("shares", 0, 'user', 200),
        ("read:shares", 2, 'server', 200),
        ("read:shares!user=USER", 3, 'user', 200),
        ("read:shares!server=SERVER", 2, 'server', 200),
        ("read:shares!server=SERVER", 2, 'user', 404),
        ("read:users:shares", 0, 'user', 403),
        ("users:shares", 1, 'server', 403),
    ],
)
async def test_share_codes_api_list(
    app, user, share_user, create_user_with_scopes, have_scopes, n_codes, level, status
):
    db = app.db
    spawner = user.spawner.orm_spawner

    _expand_scopes = partial(expand_scopes, user=user, share_with=share_user)
    requester = create_user_with_scopes(*_expand_scopes(have_scopes))

    expected_shares = []
    for i in range(n_codes):
        code = orm.ShareCode(
            spawner=spawner,
            owner=spawner.user,
            scopes=sorted(scopes.access_scopes(spawner=spawner)),
        )
        db.add(code)
        db.commit()
        expected_shares.append(f"sc_{code.id}")

    expected_shares = sorted(expected_shares)
    if level == 'user':
        path = f"/share-codes/{user.name}"
    else:
        path = f"/share-codes/{user.name}/"
    r = await api_request(app, path, name=requester.name)
    assert r.status_code == status
    if status >= 400:
        return
    share_codes = r.json()
    found_shares = []
    for share_code in share_codes["items"]:
        assert 'code' not in share_code
        assert 'id' in share_code
        assert 'server' in share_code
        found_shares.append(share_code["id"])
    found_shares = sorted(found_shares)
    assert found_shares == expected_shares


async def test_share_codes_api_list_no_such_owner(app, user):
    spawner = user.spawner.orm_spawner  # noqa
    r = await api_request(app, "/share-codes/nosuchuser")
    assert r.status_code == 404
    r = await api_request(app, "/share-codes/nosuchuser/")
    assert r.status_code == 404
    r = await api_request(app, f"/share-codes/{user.name}/nosuchserver")
    assert r.status_code == 404


@pytest.mark.parametrize(
    "have_scopes, share_scopes, status",
    [
        (None, None, 200),
        ("shares", None, 200),
        ("shares!user=other", None, 404),
        (
            "shares!server=SERVER,servers!server=SERVER",
            "read:servers!server=SERVER,access:servers!server=SERVER",
            200,
        ),
        (None, "read:servers!server=other/", 400),
        (
            "shares,access:servers",
            "admin:servers!server=SERVER",
            403,
        ),
        (None, None, 200),
    ],
)
async def test_share_codes_api_create(
    app,
    user,
    group,
    share_user,
    create_user_with_scopes,
    have_scopes,
    share_scopes,
    status,
):
    # make sure default spawner exists
    spawner = user.spawner  # noqa
    body = {}
    share_with = share_user
    _expand_scopes = partial(expand_scopes, user=user, share_with=share_with)

    expected_scopes = _expand_scopes("access:servers!server=SERVER")
    if share_scopes:
        share_scopes = _expand_scopes(share_scopes)
        expected_scopes.extend(share_scopes)
        body["scopes"] = share_scopes

    expected_scopes = sorted(set(expected_scopes))

    if have_scopes is None:
        # default: needed permissions
        have_scopes = "shares"

    requester = create_user_with_scopes(*_expand_scopes(have_scopes))

    r = await api_request(
        app,
        f"/share-codes/{user.name}/",
        method="post",
        data=json.dumps(body),
        name=requester.name,
    )
    assert r.status_code == status
    if r.status_code >= 400:
        return

    share_model = r.json()
    assert "scopes" in share_model
    assert sorted(share_model["scopes"]) == expected_scopes
    assert "code" in share_model
    assert "accept_url" in share_model
    parsed_accept_url = urlparse(share_model["accept_url"])
    accept_query = parse_qs(parsed_accept_url.query)
    assert accept_query == {"code": [share_model["code"]]}
    assert parsed_accept_url.path == url_path_join(app.base_url, "hub/accept-share")


@pytest.mark.parametrize(
    "expires_in, status",
    [
        (None, 200),
        ("notanumber", 400),
        (-1, 400),
        (60, 200),
        (525600 * 59, 200),
        (525600 * 60 + 1, 400),
    ],
)
async def test_share_codes_api_create_expires_in(
    app,
    user,
    group,
    create_user_with_scopes,
    expires_in,
    status,
):
    # make sure default spawner exists
    spawner = user.spawner  # noqa
    body = {}
    now = utcnow()
    if expires_in:
        body["expires_in"] = expires_in

    r = await api_request(
        app,
        f"/share-codes/{user.name}/",
        method="post",
        data=json.dumps(body),
    )
    assert r.status_code == status
    if r.status_code >= 400:
        return

    share_model = r.json()
    assert "expires_at" in share_model
    assert share_model["expires_at"]
    expires_at = parse_date(share_model["expires_at"])

    expected_expires_at = now + timedelta(
        seconds=expires_in or orm.ShareCode.default_expires_in
    )
    window = timedelta(seconds=60)
    assert expected_expires_at - window <= expires_at <= expected_expires_at + window

    async def get_code():
        r = await api_request(
            app,
            f"/share-codes/{user.name}/",
        )
        r.raise_for_status()
        codes = r.json()["items"]
        assert len(codes) <= 1
        if len(codes) == 1:
            return codes[0]
        else:
            return None

    code = await get_code()
    assert code

    with mock.patch(
        'jupyterhub.orm.ShareCode.now',
        staticmethod(lambda: (expires_at + timedelta(seconds=1)).replace(tzinfo=None)),
    ):
        code = await get_code()
        assert code is None


@pytest.mark.parametrize(
    "have_scopes, delete_by, status",
    [
        (None, None, 204),
        ("shares", "id=ID", 204),
        (
            "shares!server=SERVER",
            "code=CODE",
            204,
        ),
        ("shares!user=other", None, 404),
        ("read:shares", "code=CODE", 403),
        ("shares", "id=invalid", 404),
        ("shares", "id=sc_9999", 404),
        ("shares", "code=nomatch", 404),
    ],
)
async def test_share_codes_api_revoke(
    app,
    user,
    group,
    share_user,
    create_user_with_scopes,
    have_scopes,
    delete_by,
    status,
):
    db = app.db
    spawner = user.spawner.orm_spawner

    _expand_scopes = partial(expand_scopes, user=user, share_with=share_user)
    # make sure default spawner exists
    spawner = user.spawner.orm_spawner
    share_code, code = orm.ShareCode.new(
        db, spawner, scopes=list(scopes.access_scopes(spawner=spawner))
    )

    assert orm.ShareCode.find(db, code=code)
    other_share_code, other_code = orm.ShareCode.new(
        db, spawner, scopes=list(scopes.access_scopes(spawner=spawner))
    )

    if have_scopes is None:
        # default: needed permissions
        have_scopes = "shares"

    requester = create_user_with_scopes(*_expand_scopes(have_scopes))

    url = f"/share-codes/{user.name}/"
    if delete_by:
        query = delete_by.replace("CODE", code).replace("ID", f"sc_{share_code.id}")
        url = f"{url}?{query}"

    r = await api_request(
        app,
        url,
        method="delete",
        name=requester.name,
    )
    assert r.status_code == status

    # other code unaffected
    if r.status_code >= 400:
        assert orm.ShareCode.find(db, code=code)
        return
    # code has been deleted
    assert orm.ShareCode.find(db, code=code) is None
    if delete_by is None:
        assert orm.ShareCode.find(db, code=other_code) is None
    else:
        assert orm.ShareCode.find(db, code=other_code)


@pytest.mark.parametrize(
    "who, code_arg, get_status, post_status",
    [
        ("share", None, 400, 400),
        ("share", "nosuchcode", 404, 400),
        ("share", "CODE", 200, 302),
        ("self", "CODE", 403, 400),
    ],
)
async def test_accept_share_page(
    app, user, share_user, who, code_arg, get_status, post_status
):
    db = app.db
    spawner = user.spawner.orm_spawner
    orm_code, code = orm.ShareCode.new(
        db, spawner, scopes=list(scopes.access_scopes(spawner=spawner))
    )
    if who == "self":
        cookies = await app.login_user(user.name)
    else:
        cookies = await app.login_user(share_user.name)

    url = "accept-share"
    form_data = {"_xsrf": cookies['_xsrf']}
    if code_arg:
        code_arg = code_arg.replace("CODE", code)
        form_data["code"] = code_arg
        url = url + f"?code={code_arg}"

    r = await get_page(url, app, cookies=cookies)
    assert r.status_code == get_status

    # try submitting the form with the same inputs
    accept_url = public_url(app) + "hub/accept-share"
    r = await async_requests.post(
        accept_url,
        cookies=cookies,
        data=form_data,
        allow_redirects=False,
    )
    assert r.status_code == post_status
    if post_status < 400:
        assert orm_code.exchange_count == 1
        # share is accepted
        assert len(share_user.shared_with_me) == 1
        assert share_user.shared_with_me[0].spawner is spawner
    else:
        assert orm_code.exchange_count == 0
        assert not share_user.shared_with_me


@pytest.mark.parametrize(
    "running, next_url, expected_next",
    [
        (False, None, "{USER_URL}"),
        (True, None, "{USER_URL}"),
        (False, "https://example.com{BASE_URL}", "{USER_URL}"),
        (False, "{BASE_URL}hub", ""),
        (True, "{USER_URL}lab/tree/notebook.ipynb?param=5", ""),
        (False, "{USER_URL}lab/tree/notebook.ipynb?param=5", ""),
    ],
)
async def test_accept_share_page_next_url(
    app,
    user,
    share_user,
    running,
    next_url,
    expected_next,
):
    db = app.db
    spawner = user.spawner.orm_spawner
    orm_code, code = orm.ShareCode.new(
        db, spawner, scopes=list(scopes.access_scopes(spawner=spawner))
    )
    cookies = await app.login_user(share_user.name)

    if running:
        await user.spawn()
        await user.spawner.server.wait_up(http=True)
        await app.proxy.add_user(user)
    else:
        pass

    def expand_url(url):
        url = url.format(
            USER=user.name,
            USER_URL=user.server_url(""),
            BASE_URL=app.base_url,
        )
        return url

    if next_url:
        next_url = expand_url(next_url)
    if expected_next:
        expected_next = expand_url(expected_next)
    else:
        # empty: expect match
        expected_next = next_url

    url = f"accept-share?code={code}"
    form_data = {"_xsrf": cookies['_xsrf']}
    if next_url:
        url = url_concat(url, {"next": next_url})

    r = await get_page(url, app, cookies=cookies)
    assert r.status_code == 200

    page = BeautifulSoup(r.text)
    page_body = page.find("div", class_="container").get_text()
    if running:
        assert "not currently running" not in page_body
    else:
        assert "not currently running" in page_body

    # try submitting the form with the same inputs
    accept_url = r.url
    r = await async_requests.post(
        accept_url,
        cookies=cookies,
        data=form_data,
        allow_redirects=False,
    )
    assert r.status_code == 302
    target = r.headers["Location"]
    # expect absolute path redirect
    expected_next_target = urlunparse(
        urlparse(expected_next)._replace(scheme="", netloc="")
    )
    assert target == expected_next_target
    # is it worth following the redirect?



c = get_config()  # noqa


c.JupyterHub.authenticator_class = 'dummy'
c.JupyterHub.spawner_class = 'simple'

c.Authenticator.allowed_users = {"sharer", "shared-with"}

# put the current directory on sys.path for shareextension.py
from pathlib import Path

here = Path(__file__).parent.absolute()
c.Spawner.notebook_dir = str(here)

# users need sharing permissions for their own servers
c.JupyterHub.load_roles = [
    {
        "name": "user",
        "scopes": ["self", "shares!user"],
    },
]

# below are two ways to grant sharing permission to a single-user server.
# there's no reason to use both


# OAuth token should have sharing permissions,
# so JupyterLab javascript can manage shares
c.Spawner.oauth_client_allowed_scopes = ["access:servers!server", "shares!server"]

# grant $JUPYTERHUB_API_TOKEN sharing permissions
# so that _python_ code can manage shares
c.Spawner.server_token_scopes = [
    "shares!server",  # manage shares
    "servers!server",  # start/stop itself
    "users:activity!server",  # report activity
]



#!/usr/bin/env python
# Check that installed package contains everything we expect


from pathlib import Path

import jupyterhub
from jupyterhub._data import DATA_FILES_PATH

print("Checking jupyterhub._data", end=" ")
print(f"DATA_FILES_PATH={DATA_FILES_PATH}", end=" ")
DATA_FILES_PATH = Path(DATA_FILES_PATH)
assert DATA_FILES_PATH.is_dir(), DATA_FILES_PATH
for subpath in (
    "templates/spawn.html",
    "static/css/style.min.css",
    "static/components/jquery/dist/jquery.js",
    "static/js/admin-react.js",
):
    path = DATA_FILES_PATH / subpath
    assert path.is_file(), path

print("OK")

print("Checking package_data", end=" ")
jupyterhub_path = Path(jupyterhub.__file__).parent.resolve()
for subpath in (
    "alembic.ini",
    "alembic/versions/833da8570507_rbac.py",
    "event-schemas/server-actions/v1.yaml",
    "singleuser/templates/page.html",
):
    path = jupyterhub_path / subpath
    assert path.is_file(), path

print("OK")



"""Tests for PAM authentication"""

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.
import logging
from itertools import chain
from unittest import mock
from urllib.parse import urlparse

import pytest
from requests import HTTPError
from traitlets import Any, Tuple
from traitlets.config import Config

from jupyterhub import auth, crypto, orm

from .mocking import MockHub, MockPAMAuthenticator, MockStructGroup, MockStructPasswd
from .utils import add_user, async_requests, get_page, public_url


async def test_pam_auth():
    authenticator = MockPAMAuthenticator(allow_all=True)
    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'match', 'password': 'match'}
    )
    assert authorized['name'] == 'match'

    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'match', 'password': 'nomatch'}
    )
    assert authorized is None

    # Account check is on by default for increased security
    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'notallowedmatch', 'password': 'notallowedmatch'}
    )
    assert authorized is None


async def test_pam_auth_account_check_disabled():
    authenticator = MockPAMAuthenticator(allow_all=True, check_account=False)
    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'allowedmatch', 'password': 'allowedmatch'}
    )
    assert authorized['name'] == 'allowedmatch'

    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'notallowedmatch', 'password': 'notallowedmatch'}
    )
    assert authorized['name'] == 'notallowedmatch'


async def test_pam_auth_admin_groups():
    jh_users = MockStructGroup(
        'jh_users',
        ['group_admin', 'also_group_admin', 'override_admin', 'non_admin'],
        1234,
    )
    jh_admins = MockStructGroup('jh_admins', ['group_admin'], 5678)
    wheel = MockStructGroup('wheel', ['also_group_admin'], 9999)
    system_groups = [jh_users, jh_admins, wheel]

    group_admin = MockStructPasswd('group_admin', 1234)
    also_group_admin = MockStructPasswd('also_group_admin', 1234)
    override_admin = MockStructPasswd('override_admin', 1234)
    non_admin = MockStructPasswd('non_admin', 1234)
    system_users = [group_admin, also_group_admin, override_admin, non_admin]

    user_group_map = {
        'group_admin': [jh_users.gr_gid, jh_admins.gr_gid],
        'also_group_admin': [jh_users.gr_gid, wheel.gr_gid],
        'override_admin': [jh_users.gr_gid],
        'non_admin': [jh_users.gr_gid],
    }

    def getgrnam(name):
        return [x for x in system_groups if x.gr_name == name][0]

    def getpwnam(name):
        return [x for x in system_users if x.pw_name == name][0]

    def getgrouplist(name, group):
        return user_group_map[name]

    authenticator = MockPAMAuthenticator(
        admin_groups={'jh_admins', 'wheel'},
        admin_users={'override_admin'},
        allow_all=True,
    )

    # Check admin_group applies as expected
    with mock.patch.multiple(
        authenticator,
        _getgrnam=getgrnam,
        _getpwnam=getpwnam,
        _getgrouplist=getgrouplist,
    ):
        authorized = await authenticator.get_authenticated_user(
            None, {'username': 'group_admin', 'password': 'group_admin'}
        )
    assert authorized['name'] == 'group_admin'
    assert authorized['admin'] is True

    # Check multiple groups work, just in case.
    with mock.patch.multiple(
        authenticator,
        _getgrnam=getgrnam,
        _getpwnam=getpwnam,
        _getgrouplist=getgrouplist,
    ):
        authorized = await authenticator.get_authenticated_user(
            None,
            {'username': 'also_group_admin', 'password': 'also_group_admin'},
        )
    assert authorized['name'] == 'also_group_admin'
    assert authorized['admin'] is True

    # Check admin_users still applies correctly
    with mock.patch.multiple(
        authenticator,
        _getgrnam=getgrnam,
        _getpwnam=getpwnam,
        _getgrouplist=getgrouplist,
    ):
        authorized = await authenticator.get_authenticated_user(
            None,
            {'username': 'override_admin', 'password': 'override_admin'},
        )
    assert authorized['name'] == 'override_admin'
    assert authorized['admin'] is True

    # Check it doesn't admin everyone
    with mock.patch.multiple(
        authenticator,
        _getgrnam=getgrnam,
        _getpwnam=getpwnam,
        _getgrouplist=getgrouplist,
    ):
        authorized = await authenticator.get_authenticated_user(
            None, {'username': 'non_admin', 'password': 'non_admin'}
        )
    assert authorized['name'] == 'non_admin'
    assert authorized['admin'] is False


async def test_pam_auth_allowed():
    authenticator = MockPAMAuthenticator(
        allowed_users={'wash', 'kaylee'}, allow_all=False
    )

    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'kaylee', 'password': 'kaylee'}
    )
    assert authorized['name'] == 'kaylee'

    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'wash', 'password': 'nomatch'}
    )
    assert authorized is None

    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'mal', 'password': 'mal'}
    )
    assert authorized is None


async def test_pam_auth_allowed_groups():
    class TestAuthenticator(MockPAMAuthenticator):
        @staticmethod
        def _getpwnam(name):
            return MockStructPasswd(name=name)

        @staticmethod
        def _getgrnam(name):
            if name == "group":
                return MockStructGroup('grp', ['kaylee'], gid=1234)
            else:
                return None

        @staticmethod
        def _getgrouplist(username, gid):
            gids = [gid]
            if username == "kaylee":
                gids.append(1234)
            return gids

    authenticator = TestAuthenticator(allowed_groups={'group'}, allow_all=False)

    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'kaylee', 'password': 'kaylee'}
    )
    assert authorized['name'] == 'kaylee'

    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'mal', 'password': 'mal'}
    )
    assert authorized is None


async def test_pam_auth_blocked():
    # Null case compared to next case
    authenticator = MockPAMAuthenticator(allow_all=True)
    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'wash', 'password': 'wash'}
    )
    assert authorized['name'] == 'wash'

    # Blocklist basics
    authenticator = MockPAMAuthenticator(blocked_users={'wash'}, allow_all=True)
    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'wash', 'password': 'wash'}
    )
    assert authorized is None

    # User in both allowed and blocked: default deny.  Make error someday?
    authenticator = MockPAMAuthenticator(
        blocked_users={'wash'},
        allowed_users={'wash', 'kaylee'},
        allow_all=True,
    )
    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'wash', 'password': 'wash'}
    )
    assert authorized is None

    # User not in blocked set can log in
    authenticator = MockPAMAuthenticator(
        blocked_users={'wash'},
        allowed_users={'wash', 'kaylee'},
    )
    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'kaylee', 'password': 'kaylee'}
    )
    assert authorized['name'] == 'kaylee'

    # User in allowed, blocked irrelevent
    authenticator = MockPAMAuthenticator(
        blocked_users={'mal'}, allowed_users={'wash', 'kaylee'}
    )
    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'wash', 'password': 'wash'}
    )
    assert authorized['name'] == 'wash'

    # User in neither list
    authenticator = MockPAMAuthenticator(
        blocked_users={'mal'},
        allowed_users={'wash', 'kaylee'},
    )
    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'simon', 'password': 'simon'}
    )
    assert authorized is None

    authenticator = MockPAMAuthenticator(
        blocked_users=set(), allowed_users={'wash', 'kaylee'}
    )
    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'kaylee', 'password': 'kaylee'}
    )
    assert authorized['name'] == 'kaylee'


@pytest.mark.filterwarnings("ignore::DeprecationWarning")
async def test_deprecated_signatures():
    def deprecated_xlist(self, username):
        return True

    with mock.patch.multiple(
        MockPAMAuthenticator,
        check_whitelist=deprecated_xlist,
        check_group_whitelist=deprecated_xlist,
        check_blacklist=deprecated_xlist,
    ):
        deprecated_authenticator = MockPAMAuthenticator()
        authorized = await deprecated_authenticator.get_authenticated_user(
            None, {'username': 'test', 'password': 'test'}
        )

        assert authorized is not None


async def test_pam_auth_no_such_group():
    authenticator = MockPAMAuthenticator(
        allowed_groups={'nosuchcrazygroup'},
    )
    authenticator._getpwnam = MockStructPasswd
    authorized = await authenticator.get_authenticated_user(
        None, {'username': 'kaylee', 'password': 'kaylee'}
    )
    assert authorized is None


async def test_wont_add_system_user():
    user = orm.User(name='lioness4321')
    authenticator = auth.PAMAuthenticator(allowed_users={'mal'})
    authenticator.create_system_users = False
    with pytest.raises(KeyError):
        await authenticator.add_user(user)


async def test_cant_add_system_user():
    user = orm.User(name='lioness4321')
    authenticator = auth.PAMAuthenticator(allowed_users={'mal'})
    authenticator.add_user_cmd = ['jupyterhub-fake-command']
    authenticator.create_system_users = True

    class DummyFile:
        def read(self):
            return b'dummy error'

    class DummyPopen:
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs
            self.returncode = 1
            self.stdout = DummyFile()

        def wait(self):
            return

    with mock.patch.object(auth, 'Popen', DummyPopen):
        with pytest.raises(RuntimeError) as exc:
            await authenticator.add_user(user)
        assert str(exc.value) == 'Failed to create system user lioness4321: dummy error'


async def test_add_system_user():
    user = orm.User(name='lioness4321')
    authenticator = auth.PAMAuthenticator(allowed_users={'mal'})
    authenticator.create_system_users = True
    authenticator.add_user_cmd = ['echo', '/home/USERNAME']

    record = {}

    class DummyPopen:
        def __init__(self, cmd, *args, **kwargs):
            record['cmd'] = cmd
            self.returncode = 0

        def wait(self):
            return

    with mock.patch.object(auth, 'Popen', DummyPopen):
        await authenticator.add_user(user)
    assert record['cmd'] == ['echo', '/home/lioness4321', 'lioness4321']


async def test_delete_user():
    user = orm.User(name='zoe')
    a = MockPAMAuthenticator(allowed_users={'mal'})

    assert 'zoe' not in a.allowed_users
    await a.add_user(user)
    assert 'zoe' in a.allowed_users
    a.delete_user(user)
    assert 'zoe' not in a.allowed_users


def test_urls():
    a = auth.PAMAuthenticator()
    logout = a.logout_url('/base/url/')
    login = a.login_url('/base/url')
    assert logout == '/base/url/logout'
    assert login == '/base/url/login'


def test_handlers(app):
    a = auth.PAMAuthenticator()
    handlers = a.get_handlers(app)
    assert handlers[0][0] == '/login'


async def test_auth_state(app, auth_state_enabled):
    """auth_state enabled and available"""
    name = 'kiwi'
    user = add_user(app.db, app, name=name)
    assert user.encrypted_auth_state is None
    cookies = await app.login_user(name)
    auth_state = await user.get_auth_state()
    assert auth_state == app.authenticator.auth_state


async def test_auth_admin_non_admin(app):
    """admin should be passed through for non-admin users"""
    name = 'kiwi'
    user = add_user(app.db, app, name=name, admin=False)
    assert user.admin is False
    cookies = await app.login_user(name)
    assert user.admin is False


async def test_auth_admin_is_admin(app):
    """admin should be passed through for admin users"""
    # Admin user defined in MockPAMAuthenticator.
    name = 'admin'
    user = add_user(app.db, app, name=name, admin=False)
    assert user.admin is False
    cookies = await app.login_user(name)
    assert user.admin is True


async def test_auth_admin_retained_if_unset(app):
    """admin should be unchanged if authenticator doesn't return admin value"""
    name = 'kiwi'
    # Add user as admin.
    user = add_user(app.db, app, name=name, admin=True)
    assert user.admin is True
    # User should remain unchanged.
    cookies = await app.login_user(name)
    assert user.admin is True


@pytest.fixture
def auth_state_unavailable(auth_state_enabled):
    """auth_state enabled at the Authenticator level,

    but unavailable due to no crypto keys.
    """
    crypto.CryptKeeper.instance().keys = []
    yield


async def test_auth_state_disabled(app, auth_state_unavailable):
    name = 'driebus'
    user = add_user(app.db, app, name=name)
    assert user.encrypted_auth_state is None
    with pytest.raises(HTTPError):
        cookies = await app.login_user(name)
    auth_state = await user.get_auth_state()
    assert auth_state is None


async def test_normalize_names():
    a = MockPAMAuthenticator(allow_all=True)
    authorized = await a.get_authenticated_user(
        None, {'username': 'ZOE', 'password': 'ZOE'}
    )
    assert authorized['name'] == 'zoe'

    authorized = await a.get_authenticated_user(
        None, {'username': 'Glenn', 'password': 'Glenn'}
    )
    assert authorized['name'] == 'glenn'

    authorized = await a.get_authenticated_user(
        None, {'username': 'hExi', 'password': 'hExi'}
    )
    assert authorized['name'] == 'hexi'

    authorized = await a.get_authenticated_user(
        None, {'username': 'Test', 'password': 'Test'}
    )
    assert authorized['name'] == 'test'


async def test_username_map():
    a = MockPAMAuthenticator(username_map={'wash': 'alpha'}, allow_all=True)
    authorized = await a.get_authenticated_user(
        None, {'username': 'WASH', 'password': 'WASH'}
    )

    assert authorized['name'] == 'alpha'

    authorized = await a.get_authenticated_user(
        None, {'username': 'Inara', 'password': 'Inara'}
    )
    assert authorized['name'] == 'inara'


def test_validate_names():
    a = auth.PAMAuthenticator()
    assert a.validate_username('willow')
    assert a.validate_username('giles')
    assert a.validate_username('Test')
    assert a.validate_username('hExi')
    assert a.validate_username('Glenn#Smith!')
    a = auth.PAMAuthenticator(username_pattern='w.*')
    assert not a.validate_username('xander')
    assert a.validate_username('willow')


async def test_post_auth_hook():
    def test_auth_hook(authenticator, handler, authentication):
        authentication['testkey'] = 'testvalue'
        return authentication

    a = MockPAMAuthenticator(allow_all=True, post_auth_hook=test_auth_hook)

    authorized = await a.get_authenticated_user(
        None, {'username': 'test_user', 'password': 'test_user'}
    )

    assert authorized['testkey'] == 'testvalue'


class MyAuthenticator(auth.Authenticator):
    def check_whitelist(self, username, authentication=None):
        return username == "subclass-allowed"


def test_deprecated_config(caplog):
    cfg = Config()
    cfg.Authenticator.whitelist = {'user'}
    log = logging.getLogger("testlog")
    authenticator = auth.Authenticator(config=cfg, log=log)
    assert caplog.record_tuples == [
        (
            log.name,
            logging.WARNING,
            'Authenticator.whitelist is deprecated in JupyterHub 1.2, use '
            'Authenticator.allowed_users instead',
        )
    ]
    assert authenticator.allowed_users == {'user'}


def test_deprecated_methods():
    cfg = Config()
    cfg.Authenticator.whitelist = {'user'}
    authenticator = auth.Authenticator(config=cfg)

    assert authenticator.check_allowed("user")
    with pytest.deprecated_call():
        assert authenticator.check_whitelist("user")
    assert not authenticator.check_allowed("otheruser")
    with pytest.deprecated_call():
        assert not authenticator.check_whitelist("otheruser")


def test_deprecated_config_subclass():
    cfg = Config()
    cfg.MyAuthenticator.whitelist = {'user'}
    with pytest.deprecated_call():
        authenticator = MyAuthenticator(config=cfg)
    assert authenticator.allowed_users == {'user'}


def test_deprecated_methods_subclass():
    with pytest.deprecated_call():
        authenticator = MyAuthenticator()

    assert authenticator.check_allowed("subclass-allowed")
    assert authenticator.check_whitelist("subclass-allowed")
    assert not authenticator.check_allowed("otheruser")
    assert not authenticator.check_whitelist("otheruser")


async def test_nullauthenticator(app):
    with mock.patch.dict(
        app.tornado_settings, {"authenticator": auth.NullAuthenticator(parent=app)}
    ):
        r = await async_requests.get(public_url(app))
    assert urlparse(r.url).path.endswith("/hub/login")
    assert r.status_code == 403


class MockGroupsAuthenticator(auth.Authenticator):
    authenticated_groups = Any()
    refresh_groups = Any()

    manage_groups = True

    def authenticate(self, handler, data):
        return {
            "name": data["username"],
            "groups": self.authenticated_groups,
        }

    async def refresh_user(self, user, handler):
        return {
            "name": user.name,
            "groups": self.refresh_groups,
        }


@pytest.mark.parametrize(
    "authenticated_groups, refresh_groups",
    [
        (None, None),
        (["auth1"], None),
        (None, ["auth1"]),
        (["auth1"], ["auth1", "auth2"]),
        (["auth1", "auth2"], ["auth1"]),
        (["auth1", "auth2"], ["auth3"]),
        (["auth1", "auth2"], ["auth3"]),
    ],
)
async def test_auth_managed_groups(
    app, user, group, authenticated_groups, refresh_groups
):
    authenticator = MockGroupsAuthenticator(
        parent=app,
        authenticated_groups=authenticated_groups,
        refresh_groups=refresh_groups,
        allow_all=True,
    )

    user.groups.append(group)
    app.db.commit()
    before_groups = [group.name]
    if authenticated_groups is None:
        expected_authenticated_groups = before_groups
    else:
        expected_authenticated_groups = authenticated_groups
    if refresh_groups is None:
        expected_refresh_groups = expected_authenticated_groups
    else:
        expected_refresh_groups = refresh_groups

    with mock.patch.dict(app.tornado_settings, {"authenticator": authenticator}):
        cookies = await app.login_user(user.name)
        assert not app.db.dirty
        groups = sorted(g.name for g in user.groups)
        assert groups == expected_authenticated_groups

        # force refresh_user on next request
        user._auth_refreshed -= 10 + app.authenticator.auth_refresh_age
        r = await get_page('home', app, cookies=cookies, allow_redirects=False)
        assert r.status_code == 200
        assert not app.db.dirty
        groups = sorted(g.name for g in user.groups)
        assert groups == expected_refresh_groups


class MockRolesAuthenticator(auth.Authenticator):
    authenticated_roles = Any()
    refresh_roles = Any()
    initial_roles = Any()
    manage_roles = True

    def authenticate(self, handler, data):
        return {
            "name": data["username"],
            "roles": self.authenticated_roles,
        }

    async def refresh_user(self, user, handler):
        return {
            "name": user.name,
            "roles": self.refresh_roles,
        }

    async def load_managed_roles(self):
        return self.initial_roles


def role_to_dict(role):
    return {col: getattr(role, col) for col in role.__table__.columns.keys()}


@pytest.mark.parametrize(
    "initial_roles",
    [
        pytest.param(
            [{'name': 'test-role'}],
            id="should have the same effect as `load_roles` when creating a new role",
        ),
        pytest.param(
            [
                {
                    'name': 'server',
                    'users': ['test-user'],
                }
            ],
            id="should have the same effect as `load_roles` when assigning a role to a user",
        ),
        pytest.param(
            [
                {
                    'name': 'server',
                    'groups': ['test-group'],
                }
            ],
            id="should have the same effect as `load_roles` when assigning a role to a group",
        ),
    ],
)
async def test_auth_load_managed_roles(app, initial_roles):
    authenticator = MockRolesAuthenticator(
        parent=app,
        initial_roles=initial_roles,
    )

    # create the roles using `load_roles`
    hub = MockHub(load_roles=initial_roles)
    hub.init_db()
    await hub.init_role_creation()
    expected_roles = [role_to_dict(role) for role in hub.db.query(orm.Role)]

    # create the roles using authenticator's `load_managed_roles`
    hub = MockHub(load_roles=[], authenticator=authenticator)
    hub.init_db()
    await hub.init_role_creation()
    actual_roles = [role_to_dict(role) for role in hub.db.query(orm.Role)]

    # remove `managed_by_auth` from comparison as this is expected to differ
    for role in [*actual_roles, *expected_roles]:
        role.pop('managed_by_auth')

    # `load_managed_roles` should produce the same set of roles as `load_roles` does
    assert expected_roles == actual_roles


async def test_auth_load_managed_roles_handles_new_user(app):
    authenticator = MockRolesAuthenticator(
        parent=app,
        initial_roles=[
            {
                'name': 'new-role',
                'users': ['this-user-does-not-exist'],
            }
        ],
    )

    # create the roles using authenticator's `load_managed_roles`
    hub = MockHub(load_roles=[], authenticator=authenticator)
    hub.init_db()
    await hub.init_role_creation()
    roles = {role.name: role_to_dict(role) for role in hub.db.query(orm.Role)}

    assert roles['new-role']


@pytest.mark.parametrize(
    "roles_after_restart",
    [
        pytest.param(
            [{'name': 'role-a', 'scopes': ['admin:servers'], 'users': ['test-user']}],
            id="preserve role assignment if explicitly defined in `load_managed_roles()`",
        ),
        pytest.param(
            [{'name': 'role-a', 'scopes': ['admin:servers']}],
            id="preserve role assignment if `users` key is absent in `load_managed_roles()` result",
        ),
    ],
)
async def test_auth_load_managed_roles_preserves_assignments(
    app, caplog, roles_after_restart
):
    log = logging.getLogger("testlog")
    caplog.set_level(logging.INFO, logger=log.name)
    authenticator = MockRolesAuthenticator(
        parent=app,
        initial_roles=[
            {'name': 'role-a', 'scopes': ['admin:servers'], 'users': ['test-user']},
            {'name': 'role-b', 'scopes': ['admin:servers'], 'users': ['test-user']},
        ],
    )
    authenticator.reset_managed_roles_on_startup = True

    hub = MockHub(authenticator=authenticator, log=log)
    hub.init_db()

    # simulate hub startup, it should assign the roles to the user
    await hub.init_role_creation()
    await hub.init_role_assignment()
    user = orm.User.find(hub.db, 'test-user')

    user_role_names = {r.name for r in user.roles}
    assert 'role-a' in user_role_names
    assert 'role-b' in user_role_names

    message = (
        "Deleted %s stale %s role assignments previously added by an authenticator"
    )
    assert message not in {record.msg for record in caplog.records}

    # simulate hub restart after changing `load_managed_roles()` result
    authenticator.initial_roles = roles_after_restart

    await hub.init_role_assignment()
    hub.db.refresh(user)

    user_role_names = {r.name for r in user.roles}
    assert 'role-a' in user_role_names
    assert 'role-b' not in user_role_names

    record_map = {record.msg: record for record in caplog.records}
    assert message in record_map
    record = record_map[message]
    # should only delete the second assignment, hence the log should only say "1"
    assert record.getMessage().startswith('Deleted 1 stale user role assignments')


async def test_auth_load_managed_roles_preserves_roles(app, caplog):
    log = logging.getLogger("testlog")
    caplog.set_level(logging.INFO, logger=log.name)
    authenticator = MockRolesAuthenticator(
        parent=app,
        initial_roles=[
            {'name': 'role-a', 'scopes': ['admin:servers']},
            {'name': 'role-b', 'scopes': ['admin:servers']},
        ],
    )
    authenticator.reset_managed_roles_on_startup = True

    hub = MockHub(authenticator=authenticator, log=log)
    hub.init_db()

    def find_role(name):
        return hub.db.query(orm.Role).filter(orm.Role.name == name).first()

    # simulate hub startup, it should create both roles
    await hub.init_role_creation()
    await hub.init_role_assignment()
    assert find_role('role-a') and find_role('role-b')

    message = "Deleted %s stale roles previously added by an authenticator"
    assert message not in {record.msg for record in caplog.records}

    # simulate hub restart with only the first role preserved
    authenticator.initial_roles = authenticator.initial_roles[:1]

    await hub.init_role_creation()
    assert find_role('role-a') and not find_role('role-b')

    record_map = {record.msg: record for record in caplog.records}
    assert message in record_map
    record = record_map[message]
    # should only delete the second role, hence the log should only say "1"
    assert record.getMessage().startswith('Deleted 1 stale roles')


@pytest.mark.parametrize(
    "authenticated_roles",
    [
        (None),
        ([{"name": "role-1"}]),
        ([{"name": "role-2", "description": "test role 2"}]),
        ([{"name": "role-3", "scopes": ["admin:servers"]}]),
    ],
)
async def test_auth_managed_roles(app, user, role, authenticated_roles):
    authenticator = MockRolesAuthenticator(
        parent=app,
        authenticated_roles=authenticated_roles,
    )
    user.roles.append(role)
    app.db.commit()
    before_roles = [
        {
            'name': r.name,
            'description': r.description,
            'scopes': r.scopes,
            'users': r.users,
        }
        for r in user.roles
    ]

    if authenticated_roles is None:
        expected_roles = before_roles
    else:
        expected_roles = authenticated_roles

    # Check if user gets auth-managed roles
    with mock.patch.dict(app.tornado_settings, {"authenticator": authenticator}):
        await app.login_user(user.name)
        assert not app.db.dirty

        assert len(user.roles) == len(expected_roles)

        for expected_role in expected_roles:
            role = orm.Role.find(app.db, expected_role['name'])
            assert role.description == expected_role.get('description', None)
            assert len(role.scopes) == len(expected_role.get('scopes', []))


@pytest.mark.parametrize(
    "role_spec,expected",
    [
        [
            {"name": "role-with-services", "services": ["missing-service"]},
            "Could not assign the role role-with-services to services: ['missing-service']",
        ],
        [
            {"name": "role-with-users", "users": ["missing-user"]},
            "Could not assign the role role-with-users to users: ['missing-user']",
        ],
        [
            {"name": "role-with-groups", "groups": ["missing-group"]},
            "Could not assign the role role-with-groups to groups: ['missing-group']",
        ],
    ],
)
async def test_auth_manage_roles_warns_about_unknown_entities(
    app, user, role_spec, expected
):
    # Add the current user to test that non-missing entities are not included in the warning
    role_spec['users'] = [*role_spec.get('users', []), user.name]
    # Add a scope to silence "Role will have no scopes" warning
    role_spec['scopes'] = ['admin:servers']

    authenticator = MockRolesAuthenticator(parent=app, authenticated_roles=[role_spec])

    logs = []

    def log_mock(template, *args):
        logs.append(template.format(args))

    with mock.patch.dict(app.tornado_settings, {"authenticator": authenticator}):
        with mock.patch.object(user.log, 'warning', new=log_mock):
            await app.login_user(user.name)
            assert len(logs) == 1
            assert expected in logs[0]


async def test_auth_manage_roles_strips_user_of_old_roles(app, user, role):
    authenticator = MockRolesAuthenticator(parent=app, authenticated_roles=[])
    user.roles.append(role)
    app.db.commit()
    assert [role.name for role in user.roles] == ['user', role.name]

    with mock.patch.dict(app.tornado_settings, {"authenticator": authenticator}):
        await app.login_user(user.name)
        assert not app.db.dirty
        assert [role.name for role in user.roles] == []


async def test_auth_manage_roles_grants_new_roles(app, user, role):
    authenticator = MockRolesAuthenticator(
        parent=app, authenticated_roles=[{'name': 'user'}, role_to_dict(role)]
    )
    assert [role.name for role in user.roles] == ['user']

    with mock.patch.dict(app.tornado_settings, {"authenticator": authenticator}):
        await app.login_user(user.name)
        assert not app.db.dirty
        assert [role.name for role in user.roles] == ['user', role.name]


async def test_auth_manage_roles_removes_no_longer_used_roles(app, user):
    count_roles = (
        app.db.query(orm.Role).filter(orm.Role.name == 'new-managed-role').count
    )
    authenticator = MockRolesAuthenticator(
        parent=app, authenticated_roles=[{'name': 'new-managed-role'}]
    )
    # this role should not exist yet
    assert count_roles() == 0

    with mock.patch.dict(app.tornado_settings, {"authenticator": authenticator}):
        await app.login_user(user.name)
        assert count_roles() == 1

        authenticator.authenticated_roles = []

        await app.login_user(user.name)
        assert count_roles() == 0


async def test_auth_manage_roles_does_not_remove_stripped_roles_if_used(
    app, user, group
):
    find_role = app.db.query(orm.Role).filter(orm.Role.name == 'new-managed-role').first
    authenticator = MockRolesAuthenticator(
        parent=app, authenticated_roles=[{'name': 'new-managed-role'}]
    )
    # this role should not exist yet
    assert find_role() is None

    with mock.patch.dict(app.tornado_settings, {"authenticator": authenticator}):
        await app.login_user(user.name)
        role = find_role()
        assert role

        # assign a group to the role
        role.groups = [group]
        app.db.commit()

        authenticator.authenticated_roles = []

        await app.login_user(user.name)
        assert find_role()


async def test_auth_manage_roles_does_not_remove_non_managed_roles(app, user, role):
    count_roles = app.db.query(orm.Role).filter(orm.Role.name == role.name).count
    authenticator = MockRolesAuthenticator(
        parent=app, authenticated_roles=[role_to_dict(role)]
    )
    # this role is created by the fixture and should already exist
    assert count_roles() == 1

    with mock.patch.dict(app.tornado_settings, {"authenticator": authenticator}):
        await app.login_user(user.name)

        authenticator.authenticated_roles = []

        await app.login_user(user.name)
        # should not delete the role as this role was not created by the authenticator
        assert count_roles() == 1


async def test_auth_manage_roles_marks_new_role_as_managed(app, user):
    authenticator = MockRolesAuthenticator(
        parent=app, authenticated_roles=[{'name': 'new-role'}]
    )

    with mock.patch.dict(app.tornado_settings, {"authenticator": authenticator}):
        await app.login_user(user.name)
        assert not app.db.dirty
        assert user.roles[0].managed_by_auth


async def test_auth_manage_roles_marks_new_assignment_as_managed(app, user, role):
    authenticator = MockRolesAuthenticator(
        parent=app, authenticated_roles=[role_to_dict(role)]
    )

    with mock.patch.dict(app.tornado_settings, {"authenticator": authenticator}):
        await app.login_user(user.name)
        assert not app.db.dirty
        UserRoleMap = orm._role_associations['user']
        association = (
            app.db.query(UserRoleMap)
            .filter((UserRoleMap.role_id == role.id) & (UserRoleMap.user_id == user.id))
            .one()
        )
        assert association.managed_by_auth


@pytest.mark.parametrize(
    "role_spec,expected",
    [
        pytest.param(
            {'name': 'role-a'},
            'Test description',
            id="should keep the original role description",
        ),
        pytest.param(
            {'name': 'role-b', 'description': 'New description'},
            'New description',
            id="should update the role description",
        ),
    ],
)
async def test_auth_manage_roles_description_handling(app, user, role_spec, expected):
    authenticator = MockRolesAuthenticator(parent=app, authenticated_roles=[role_spec])
    name = role_spec['name']
    role = orm.Role(name=name, description='Test description')
    user.roles.append(role)
    app.db.commit()

    with mock.patch.dict(app.tornado_settings, {"authenticator": authenticator}):
        await app.login_user(user.name)
        assert not app.db.dirty
        roles = {role.name: role for role in user.roles}
        assert roles[name].description == expected


@pytest.mark.parametrize(
    "allowed_users,  allow_existing_users",
    [
        ('specified', True),
        ('', False),
    ],
)
async def test_allow_defaults(app, user, allowed_users, allow_existing_users):
    if allowed_users:
        allowed_users = set(allowed_users.split(','))
    else:
        allowed_users = set()
    authenticator = auth.Authenticator(allowed_users=allowed_users)
    authenticator.authenticate = lambda handler, data: data["username"]
    assert authenticator.allow_all is False
    assert authenticator.allow_existing_users == allow_existing_users

    # user was already in the database
    # this happens during hub startup
    authenticator.add_user(user)
    if allowed_users:
        assert user.name in authenticator.allowed_users
    else:
        authenticator.allowed_users == set()

    specified_allowed = await authenticator.get_authenticated_user(
        None, {"username": "specified"}
    )
    if "specified" in allowed_users:
        assert specified_allowed is not None
    else:
        assert specified_allowed is None

    existing_allowed = await authenticator.get_authenticated_user(
        None, {"username": user.name}
    )
    if allow_existing_users:
        assert existing_allowed is not None
    else:
        assert existing_allowed is None


@pytest.mark.parametrize("allow_all", [None, True, False])
@pytest.mark.parametrize("allow_existing_users", [None, True, False])
@pytest.mark.parametrize("allowed_users", ["existing", ""])
def test_allow_existing_users(
    app, user, allowed_users, allow_all, allow_existing_users
):
    if allowed_users:
        allowed_users = set(allowed_users.split(','))
    else:
        allowed_users = set()
    authenticator = auth.Authenticator(
        allowed_users=allowed_users,
    )
    if allow_all is None:
        # default allow_all
        allow_all = authenticator.allow_all
    else:
        authenticator.allow_all = allow_all
    if allow_existing_users is None:
        # default allow_all
        allow_existing_users = authenticator.allow_existing_users
    else:
        authenticator.allow_existing_users = allow_existing_users

    # first, nobody in the database
    assert authenticator.check_allowed("newuser") == allow_all

    # user was already in the database
    # this happens during hub startup
    authenticator.add_user(user)
    if allow_existing_users or allow_all:
        assert authenticator.check_allowed(user.name)
    else:
        assert not authenticator.check_allowed(user.name)
    for username in allowed_users:
        assert authenticator.check_allowed(username)

    assert authenticator.check_allowed("newuser") == allow_all


@pytest.mark.parametrize("allow_all", [True, False])
@pytest.mark.parametrize("allow_existing_users", [True, False])
def test_allow_existing_users_first_time(user, allow_all, allow_existing_users):
    # make sure that calling add_user doesn't change results
    authenticator = auth.Authenticator(
        allow_all=allow_all,
        allow_existing_users=allow_existing_users,
    )
    allowed_before_one = authenticator.check_allowed(user.name)
    allowed_before_two = authenticator.check_allowed("newuser")
    # add_user is called after successful login
    # it shouldn't change results (e.g. by switching .allowed_users from empty to non-empty)
    if allowed_before_one:
        authenticator.add_user(user)
    assert authenticator.check_allowed(user.name) == allowed_before_one
    assert authenticator.check_allowed("newuser") == allowed_before_two


class AllowAllIgnoringAuthenticator(auth.Authenticator):
    """Test authenticator with custom check_allowed

    not updated for allow_all, allow_existing_users

    Make sure new config doesn't break backward-compatibility
    or grant unintended access for Authenticators written before JupyterHub 5.
    """

    allowed_letters = Tuple(config=True, help="Initial letters to allow")

    def authenticate(self, handler, data):
        return {"name": data["username"]}

    def check_allowed(self, username, auth=None):
        if not self.allowed_users and not self.allowed_letters:
            # this subclass doesn't know about the JupyterHub 5 allow_all config
            # no allow config, allow all!
            return True
        if self.allowed_users and username in self.allowed_users:
            return True
        if self.allowed_letters and username.startswith(self.allowed_letters):
            return True
        return False


# allow_all is not recognized by Authenticator subclass
# make sure it doesn't make anything more permissive, at least
@pytest.mark.parametrize("allow_all", [True, False])
@pytest.mark.parametrize(
    "allowed_users, allowed_letters, allow_existing_users, allowed, not_allowed",
    [
        ("", "", None, "anyone,should-be,allowed,existing", ""),
        ("", "a,b", None, "alice,bebe", "existing,other"),
        ("", "a,b", False, "alice,bebe", "existing,other"),
        ("", "a,b", True, "alice,bebe,existing", "other"),
        ("specified", "a,b", None, "specified,alice,bebe,existing", "other"),
        ("specified", "a,b", False, "specified,alice,bebe", "existing,other"),
        ("specified", "a,b", True, "specified,alice,bebe,existing", "other"),
    ],
)
async def test_authenticator_without_allow_all(
    app,
    allowed_users,
    allowed_letters,
    allow_existing_users,
    allowed,
    not_allowed,
    allow_all,
):
    kwargs = {}
    if allow_all is not None:
        kwargs["allow_all"] = allow_all
    if allow_existing_users is not None:
        kwargs["allow_existing_users"] = allow_existing_users
    if allowed_users:
        kwargs["allowed_users"] = set(allowed_users.split(','))
    if allowed_letters:
        kwargs["allowed_letters"] = tuple(allowed_letters.split(','))

    authenticator = AllowAllIgnoringAuthenticator(**kwargs)

    # load one user from db
    existing_user = add_user(app.db, app, name="existing")
    authenticator.add_user(existing_user)

    if allowed:
        allowed = allowed.split(",")
    if not_allowed:
        not_allowed = not_allowed.split(",")

    expected_allowed = sorted(allowed)
    expected_not_allowed = sorted(not_allowed)
    to_check = list(chain(expected_allowed, expected_not_allowed))
    if allow_all:
        expected_allowed = to_check
        expected_not_allowed = []

    are_allowed = []
    are_not_allowed = []
    for username in to_check:
        if await authenticator.get_authenticated_user(None, {"username": username}):
            are_allowed.append(username)
        else:
            are_not_allowed.append(username)

    assert are_allowed == expected_allowed
    assert are_not_allowed == expected_not_allowed



import sys

c = get_config()  # noqa

# To run the announcement service managed by the hub, add this.

port = 9999
c.JupyterHub.services = [
    {
        'name': 'announcement',
        'url': f'http://127.0.0.1:{port}',
        'command': [
            sys.executable,
            "-m",
            "announcement",
            '--port',
            str(port),
        ],
    }
]

# The announcements need to get on the templates somehow, see page.html
# for an example of how to do this.

c.JupyterHub.template_paths = ["templates"]

c.Authenticator.allowed_users = {"announcer", "otheruser"}

# grant the 'announcer' permission to access the announcement service
c.JupyterHub.load_roles = [
    {
        "name": "announcers",
        "users": ["announcer"],
        "scopes": ["access:services!service=announcement"],
    }
]

# dummy spawner and authenticator for testing, don't actually use these!
c.JupyterHub.authenticator_class = 'dummy'
c.JupyterHub.spawner_class = 'simple'
c.JupyterHub.ip = '127.0.0.1'  # let's just run on localhost while dummy auth is enabled
